@@ -1,20 +1,20 @@
 ##############################################################################
-# Kernel discriminant analysis
+## Kernel discriminant analysis
 ###############################################################################
 
 ###############################################################################
-# Find bandwidths for each class in training set, for 2- to 6-dim 
-#
-# Parameters
-# x - data values
-# group - group variable
-# bw - type of bandwidth selector
-# nstage, pilot, pre - parameters for plugin bandwidths
-# diag - FALSE - use full b/w matrices
-#      - TRUE - use diag b/w matrices
-#
-# Returns
-# Matrix of bandwidths for each group in training set
+## Find bandwidths for each class in training set, for 2- to 6-dim 
+##
+## Parameters
+## x - data values
+## group - group variable
+## bw - type of bandwidth selector
+## nstage, pilot, pre - parameters for plugin bandwidths
+## diag - FALSE - use full b/w matrices
+##      - TRUE - use diag b/w matrices
+##
+## Returns
+## Matrix of bandwidths for each group in training set
 ###############################################################################
 
 hkda <- function(x, x.group, bw="plugin", ...)
@@ -89,17 +89,17 @@ Hkda.diag <- function(x, x.group, bw="plugin", ...)
 
 
 ###############################################################################
-# Compares true group classification with an estimated one
-#
-# Parameters
-# group - true group variable
-# est.group - estimated group variable
-#
-# Returns
-# List with components
-# comp - cross-classification table of groupings - true groups are the rows,
-#        estimated groups are the columns
-# error - total mis-classification rate
+## Compares true group classification with an estimated one
+##
+## Parameters
+## group - true group variable
+## est.group - estimated group variable
+##
+## Returns
+## List with components
+## comp - cross-classification table of groupings - true groups are the rows,
+##        estimated groups are the columns
+## error - total mis-classification rate
 ###############################################################################
 
 compare <- function(x.group, est.group, by.group=FALSE)
@@ -142,21 +142,21 @@ compare <- function(x.group, est.group, by.group=FALSE)
 }
 
 ###############################################################################
-# Computes cross-validated misclassification rates (for use when test data is
-# not independent of training data) for KDA
-#
-# Parameters
-# x - training data
-# x.group - group variable for x
-# y - data values to be classified
-# Hs - bandwidth matrices
-# prior.prob - prior probabilities
-#
-# Returns
-# List with components
-# comp - cross-classification table of groupings - true groups are the rows,
-#        estimated groups are the columns
-# error - total mis-classification rate
+## Computes cross-validated misclassification rates (for use when test data is
+## not independent of training data) for KDA
+##
+## Parameters
+## x - training data
+## x.group - group variable for x
+## y - data values to be classified
+## Hs - bandwidth matrices
+## prior.prob - prior probabilities
+##
+## Returns
+## List with components
+## comp - cross-classification table of groupings - true groups are the rows,
+##        estimated groups are the columns
+## error - total mis-classification rate
 ###############################################################################
 
 compare.kda.cv <- function(x, x.group, bw="plugin", prior.prob=NULL, Hstart, by.group=FALSE, verbose=FALSE, recompute=FALSE, ...)
@@ -194,7 +194,6 @@ compare.kda.cv <- function(x, x.group, bw="plugin", prior.prob=NULL, Hstart, by.
     return(compare(x.group, kda.cv.gr, by.group=by.group)) 
   }
 
-
   ## multi-dimensional   
   n <- nrow(x)
   d <- ncol(x)
@@ -251,7 +250,7 @@ compare.kda.cv <- function(x, x.group, bw="plugin", prior.prob=NULL, Hstart, by.
 }
 
 ###############################################################################
-### Same as compare.kda.cv except uses diagonal b/w matrices
+## Same as compare.kda.cv except uses diagonal b/w matrices
 ###############################################################################
 
 compare.kda.diag.cv <- function(x, x.group, bw="plugin", prior.prob=NULL,
@@ -295,30 +294,28 @@ compare.kda.diag.cv <- function(x, x.group, bw="plugin", prior.prob=NULL,
 }
 
 
-
-
 ###############################################################################
-# KDEs of individual densities for KDA - 1- to 3-dim
-#
-# Parameters
-# x - data values
-# group - group variable
-# Hs - bandwidth matrices
-#
-# Returns
-# List with components (class dade)
-# x - list of data values
-# eval.points - evaluation points of dnesity estimate
-# estimate - list of density estimate
-# H - list of bandwidth matrices
+## KDEs of individual densities for KDA - 1- to 3-dim
+##
+## Parameters
+## x - data values
+## group - group variable
+## Hs - bandwidth matrices
+##
+## Returns
+## List with components (class dade)
+## x - list of data values
+## eval.points - evaluation points of dnesity estimate
+## estimate - list of density estimate
+## H - list of bandwidth matrices
 ##############################################################################
 
 kda <- function(x, x.group, Hs, hs, prior.prob=NULL, gridsize, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, kde.flag=TRUE)
 {
     if (missing(eval.points)) eval.points <- x
-    ##if (is.factor(x.group)) gr <- levels(x.group)
-    gr <- sort(unique(x.group))
-    
+    if (!is.factor(x.group)) x.group <- factor(x.group) 
+    ##gr <- sort(unique(x.group))
+    gr <- levels(x.group)
     m <- length(gr)
     
     ## default values 
@@ -363,19 +360,19 @@ kda <- function(x, x.group, Hs, hs, prior.prob=NULL, gridsize, xmin, xmax, supp=
     
     ## Assign y according largest weighted density value 
     disc.gr.temp <- apply(fhat.wt, 1, which.max)
-    ##disc.gr <- factor(disc.gr.temp, levels=gr[order(unique(x.group))])
     disc.gr <- as.factor(gr[disc.gr.temp])
     if (is.numeric(gr)) disc.gr <- as.numeric(levels(disc.gr))[disc.gr]
     
     if (kde.flag) fhat.list$x.group.estimate <- disc.gr
     else fhat.list <- disc.gr
+    fhat.list$type <- "kda"
     
     return(fhat.list)
 }
 
 kda.1d <- function(x, x.group, hs, prior.prob, gridsize, supp, eval.points, binned, bgridsize, xmin, xmax, w, compute.cont, approx.cont)
 {
-    gr <- sort(unique(x.group))
+    gr <- levels(x.group)
     m <- length(gr)
     d <- 1
     hmax <- max(hs)
@@ -409,7 +406,7 @@ kda.1d <- function(x, x.group, hs, prior.prob, gridsize, supp, eval.points, binn
             fhat.list$cont <- c(fhat.list$cont, list(contlev))
         }
     }
-    
+    fhat.list$names <- parse.name(x)  ## add variable names
     fhat.list$binned <- binned
     fhat.list$gridded <- fhat.temp$gridded
     
@@ -433,7 +430,8 @@ kda.1d <- function(x, x.group, hs, prior.prob, gridsize, supp, eval.points, binn
 kda.nd <- function(x, x.group, Hs, prior.prob, gridsize, supp, eval.points, binned, bgridsize, xmin, xmax, w, compute.cont, approx.cont)
 {
   if (is.data.frame(x)) x <- as.matrix(x)
-  gr <- sort(unique(x.group))
+  gr <- levels(x.group)
+  ##gr <- sort(unique(x.group))
   m <- length(gr)
   d <- ncol(x)
   
@@ -481,6 +479,7 @@ kda.nd <- function(x, x.group, Hs, prior.prob, gridsize, supp, eval.points, binn
     }
   }
 
+  fhat.list$names <- parse.name(x)  ## add variable names
   fhat.list$binned <- binned
   fhat.list$gridded <- fhat.temp$gridded
 
@@ -494,22 +493,18 @@ kda.nd <- function(x, x.group, Hs, prior.prob, gridsize, supp, eval.points, binn
   else
     fhat.list$prior.prob <- prior.prob
 
-  ##pr <- rep(0, length(gr))
-  ##for (j in 1:length(gr)) pr[j] <- length(which(x.group==gr[j]))
-  ##pr <- pr/nrow(x)
-  ##fhat.list$prior.prob <- pr
   fhat.list$x.group <- x.group
   
   class(fhat.list) <- "kda"
   return (fhat.list)
 }
 
-
   
 ##############################################################################
-## Contour method for kda objects
+## S3 methods for kda objects
 ##############################################################################
 
+## contourLevel method
 contourLevels.kda <- function(x, prob, cont, nlevels=5, approx=TRUE,...) 
 {
   fhat <- x
@@ -526,6 +521,7 @@ contourLevels.kda <- function(x, prob, cont, nlevels=5, approx=TRUE,...)
   return(hts) 
 }
 
+## predict method
 
 predict.kda <- function(object, ..., x)
 {
@@ -543,21 +539,10 @@ predict.kda <- function(object, ..., x)
   return(est.group)
 }
 
-
-
-
-##############################################################################
-# Plot KDE of individual densities and partition - only for 2-dim
-#
-# Parameters
-# fhat - output from `kda.kde'
-# y - data points (separate from training data inside fhat)
-# y.group - data group labels
-# prior.prob - vector of prior probabilities
-# disp - "part" - plot partition
-#      - "" - don't plot partition
-##############################################################################
-
+## plot method
+## fhat - output from `kda.kde'
+## y - data points (separate from training data inside fhat)
+## y.group - data group labels
 
 plot.kda <- function(x, y, y.group, ...) 
 {
@@ -581,11 +566,11 @@ plot.kda <- function(x, y, y.group, ...)
 }
 
 
-plotkda.1d <- function(x, y, y.group, prior.prob=NULL, xlim, ylim, xlab="x", ylab="Weighted density function", drawpoints=FALSE, col, col.part, col.pt, lty, jitter=TRUE, rugsize, ...)
+plotkda.1d <- function(x, y, y.group, prior.prob=NULL, xlim, ylim, xlab, ylab="Weighted density function", drawpoints=FALSE, col, col.fun, col.part, col.pt, lty, jitter=TRUE, rugsize, add=FALSE, alpha=1, ...)
 { 
   fhat <- x
   m <- length(fhat$x)
-  
+  if (missing(xlab)) xlab <- fhat$names
   if (is.null(prior.prob))
     prior.prob <- fhat$prior.prob
   
@@ -601,31 +586,28 @@ plotkda.1d <- function(x, y, y.group, prior.prob=NULL, xlim, ylim, xlab="x", yla
   if (missing(ylim)) ylim <- range(weighted.fhat)
   if (missing(lty)) lty <- rep(1, m)
   if (length(lty) < m) lty <- rep(lty, m)
-  if (missing(col)) col <- 1:m
+  if (missing(col.fun)) col.fun <- function(n) {hcl.colors(n, palette="Dark2")}
+  if (missing(col)) col <- col.fun(m)
   if (length(col) < m) col <- rep(col, m)
-  if (missing(col.pt)) col.pt <- col
-  if (length(col.pt) < m) col.pt <- rep(col.pt, m)
-  if (missing(col.part)) col.part <- col
-  if (length(col.part) < m) col.part <- rep(col.part, m)
+  if (missing(col.part)) col.part <- plot3D::alpha.col(col, alpha=alpha) 
+  if (missing(col.pt)) col.pt <- col.fun(m)
+  if (length(col.pt)==1) col.pt <- rep(col.pt, m)
   
   ## plot each training group's KDE in separate colour and line type 
-  plot(fhat$eval.points, weighted.fhat[,1], type="l", xlab=xlab, ylab=ylab, xlim=xlim, ylim=ylim, lty=lty[1], col=col[1], ...)
+  if (!add) plot(fhat$eval.points, weighted.fhat[,1], type="l", xlab=xlab, ylab=ylab, xlim=xlim, ylim=ylim, lty=lty[1], col=col[1], ...)
   
   if (m > 1)
-    for (j in 2:m)
-      lines(fhat$eval.points, weighted.fhat[,j], lty=lty[j], col=col[j], ...)
+    for (j in 2:m) lines(fhat$eval.points, weighted.fhat[,j], lty=lty[j], col=col[j], ...)
 
   ydata <- seq(min(fhat$eval.points), max(fhat$eval.points), length=401)
-  ydata.gr <- unique(fhat$x.group)[apply(weighted.fhat,1, which.max)] 
+  x.gr <- 1:m
+  ydata.gr <- x.gr[apply(weighted.fhat,1, which.max)] 
 
   ## draw partition class as rug-like plot
- 
   plot.lim <- par()$usr
   if (missing(rugsize)) rugsize <- abs(plot.lim[4]-plot.lim[3])/50
-  for (j in 1:m)
-  {
-    image(ydata, c(plot.lim[3], plot.lim[3]+rugsize), cbind(as.numeric(ydata.gr), as.numeric(ydata.gr)), level=0.5+(0:length(levels(fhat$x.group))), add=TRUE, col=col.part, ...)
-  } 
+
+  image(ydata, c(plot.lim[3], plot.lim[3]+rugsize), cbind(as.numeric(ydata.gr), as.numeric(ydata.gr)), breaks=0.5+(0:m), add=TRUE, col=col.part, ...)  
   
   for (j in 1:m)
   {  
@@ -654,14 +636,12 @@ plotkda.1d <- function(x, y, y.group, prior.prob=NULL, xlim, ylim, xlab="x", yla
   }   
 }
 
-
-plotkda.2d <- function(x, y, y.group, prior.prob=NULL, 
+plotkda.2d <- function(x, y, y.group, prior.prob=NULL, display.part="filled.contour",
     cont=c(25,50,75), abs.cont, approx.cont=TRUE, xlim, ylim, xlab, ylab,
-    drawpoints=FALSE, drawlabels=TRUE, cex=1, pch, lty, col, col.part, col.pt, display.part="filled.contour", ...)
+    drawpoints=FALSE, drawlabels=TRUE, cex=1, pch, lty, part=TRUE, col, col.fun, col.part, col.pt, alpha=1, lwd=1, lwd.part=0, labcex=1, add=FALSE, ...)
 { 
   fhat <- x
   m <- length(fhat$x)
-  
   xtemp <- numeric()
   for (j in 1:m) xtemp <- rbind(xtemp, fhat$x[[j]]) 
   if (missing(xlim)) xlim <- range(xtemp[,1])
@@ -669,18 +649,14 @@ plotkda.2d <- function(x, y, y.group, prior.prob=NULL,
   if (missing(pch)) pch <- 1:m
   if (missing(lty)) lty <- rep(1, m)
   if (length(lty) < m) lty <- rep(lty, m)
-  if (missing(col)) col <- 1:m
+  if (missing(col.fun)) col.fun <- function(n) {hcl.colors(n, palette="Dark2", alpha=1)}
+  if (missing(col)) col <- col.fun(m)
   if (length(col) < m) col <- rep(col, m)
-  if (missing(col.part))
-      if (display.part=="slice") col.part <- 1:3
-      else col.part <- grey.colors(m, start=0.7, end=1, alpha=0.5)
- 
-  if (missing(col.pt))
-    if (missing(y.group)) col.pt <- rep("blue", m)
-    else col.pt <- 1:m
+  if (missing(col.part)) col.part <- plot3D::alpha.col(col, alpha=alpha) 
+  if (missing(col.pt)) col.pt <- col.fun(m)
   if (length(col.pt)==1) col.pt <- rep(col.pt, m)
-              
-  x.names <- colnames(fhat$x[[1]]) 
+  
+  x.names <- fhat$names ##x.names <- colnames(fhat$x[[1]]) 
   if (!is.null(x.names))
   {
     if (missing(xlab)) xlab <- x.names[1]
@@ -691,7 +667,6 @@ plotkda.2d <- function(x, y, y.group, prior.prob=NULL,
     xlab="x"
     ylab="y"
   }
-
   if (is.null(prior.prob)) prior.prob <- fhat$prior.prob
 
   if (m != length(prior.prob))
@@ -700,11 +675,13 @@ plotkda.2d <- function(x, y, y.group, prior.prob=NULL,
     stop("Sum of weights not equal to 1")
 
   ## set up plot
-  if (missing(y)) 
-    plot(fhat$x[[1]], type="n", xlab=xlab, ylab=ylab, xlim=xlim, ylim=ylim, ...)
-  else
-    plot(y, type="n", xlab=xlab, ylab=ylab, xlim=xlim, ylim=ylim, ...)
-  
+  if (!add)
+  {
+  	if (missing(y)) 
+   		plot(fhat$x[[1]], type="n", xlab=xlab, ylab=ylab, xlim=xlim, ylim=ylim, ...)
+  	else
+    	plot(y, type="n", xlab=xlab, ylab=ylab, xlim=xlim, ylim=ylim, ...)
+  }
 
   ## set up common grid for all densities 
   class.grid <- array(0, dim=dim(fhat$estimate[[1]]))
@@ -717,7 +694,6 @@ plotkda.2d <- function(x, y, y.group, prior.prob=NULL,
   }
 
   ## draw partition
-
   fhat.part <- fhat
   fhat.part$estimate <- class.grid  
   fhat.part$H <- fhat$H[[1]]
@@ -725,11 +701,10 @@ plotkda.2d <- function(x, y, y.group, prior.prob=NULL,
   fhat.part$w <- fhat$w[[1]]
   fhat.part$cont <- fhat$cont[[1]]
   class(fhat.part) <- "kde.part"
-
-  plot(fhat.part, col=col.part, add=TRUE, display=display.part, drawlabels=drawlabels, ...)
+  
+  if (part) plot(fhat.part, col=col.part, add=TRUE, display=display.part, lwd=lwd.part, alpha=alpha, ...)
   
   ## common contour levels removed from >= v1.5.3 
-
   if (missing(abs.cont))
   {
     hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
@@ -751,14 +726,14 @@ plotkda.2d <- function(x, y, y.group, prior.prob=NULL,
         scale <- cont[i]/hts[[j]][i]
         contour(fhat$eval.points[[1]], fhat$eval.points[[2]], 
                 fhat$estimate[[j]]*scale, level=hts[[j]][i]*scale, add=TRUE, 
-                drawlabels=drawlabels, lty=lty[j], col=col[j],
+                drawlabels=drawlabels, lty=lty[j], col=col[j], lwd=lwd, labcex=labcex, 
                 ...)
       }
       else
       {
         contour(fhat$eval.points[[1]], fhat$eval.points[[2]], 
                 fhat$estimate[[j]], level=hts[i], add=TRUE, 
-                drawlabels=drawlabels, lty=lty[j], col=col[j],
+                drawlabels=drawlabels, lty=lty[j], col=col[j], lwd=lwd, labcex=labcex, 
                 ...)
       }
     }
@@ -770,11 +745,14 @@ plotkda.2d <- function(x, y, y.group, prior.prob=NULL,
     if (drawpoints)
     {
       if (missing(y))
-        points(fhat$x[[j]], pch=pch[j], col=col.pt[1], cex=cex)
+        points(fhat$x[[j]], pch=pch[j], col=col.pt[j], cex=cex)
       else 
       {
         if (missing(y.group))
-          points(y, col=col.pt[1], cex=cex)
+        {  
+        	points(y, col=col.pt[1], cex=cex)
+        	j <- m+1
+        }
         else
           points(y[y.group==levels(y.group)[j],], pch=pch[j], col=col.pt[j], cex=cex) 
       }
@@ -782,12 +760,8 @@ plotkda.2d <- function(x, y, y.group, prior.prob=NULL,
   }   
 }
 
-plotkda.3d <- function(x, y, y.group, prior.prob=NULL, cont=c(25,50,75), abs.cont, approx.cont=TRUE, colors, alpha=0.5, alphavec, xlab, ylab, zlab, drawpoints=FALSE, size=3, col.pt="blue", add=FALSE, ...)
+plotkda.3d <- function(x, y, y.group, prior.prob=NULL, display="plot3D", cont=c(25,50,75), abs.cont, approx.cont=TRUE, colors, col, col.fun, col.pt, alpha=0.5, alphavec, xlab, ylab, zlab, drawpoints=FALSE, size=3, cex=1, pch, theta=-30, phi=40, d=4, ticktype="detailed", bty="f", add=FALSE, ...)
 {
-    ## suggestions from Viktor Petukhov 08/03/2018
-    if (!requireNamespace("rgl", quietly=TRUE)) stop("Install the rgl package as it is required.", call.=FALSE)
-    if (!requireNamespace("misc3d", quietly=TRUE)) stop("Install the misc3d package as it is required.", call.=FALSE)
-
     fhat <- x
     
     m <- length(fhat$x) 
@@ -797,8 +771,15 @@ plotkda.3d <- function(x, y, y.group, prior.prob=NULL, cont=c(25,50,75), abs.con
     if (!(identical(all.equal(sum(prior.prob), 1), TRUE)))  
         stop("Sum of prior weights not equal to 1")
     
+    if (missing(col.fun)) col.fun <- function(n) {hcl.colors(n, palette="Dark2")}
+    if (missing(col)) col <- col.fun(m)
+    if (length(col) < m) col <- rep(col, m)
+    if (missing(col.pt)) col.pt <- col.fun(m)
+    if (length(col.pt)==1) col.pt <- rep(col.pt, m)
+    if (missing(pch)) pch <- 1:m
+    if (length(pch)==1) pch <- rep(pch, m)
+   
     x.names <- colnames(fhat$x[[1]])
-    
     if (missing(xlab)) if (is.null(x.names)) xlab <- "x" else xlab <- x.names[1]
     if (missing(ylab)) if (is.null(x.names)) ylab <- "y" else ylab <- x.names[2]
     if (missing(zlab)) if (is.null(x.names)) zlab <- "z" else zlab <- x.names[3]
@@ -807,7 +788,6 @@ plotkda.3d <- function(x, y, y.group, prior.prob=NULL, cont=c(25,50,75), abs.con
     for (j in 1:m) xx <- rbind(xx, fhat$x[[j]])
     
     ## common contour levels removed from >= v1.5.3 
-    
     if (missing(abs.cont))
     {
         hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
@@ -819,41 +799,80 @@ plotkda.3d <- function(x, y, y.group, prior.prob=NULL, cont=c(25,50,75), abs.con
         nhts <- length(hts)
     }
     
-    if (missing(alphavec)) alphavec <- seq(0.1,0.3,length=nhts)
-    if (missing(colors)) colors <- rainbow(m)
-    if (missing(col.pt))
-        if (missing(y.group)) col.pt <- rep("blue", m)
-        else col.pt <- 1:m
-    if (length(col.pt)==1) col.pt <- rep(col.pt, m)
+    if (missing(alphavec)) alphavec <- seq(0.05,0.2,length=nhts)
     
-    xtemp <- numeric(); for (i in 1:length(fhat$x)) xtemp <- rbind(xtemp, fhat$x[[i]])
-    rgl::plot3d(x=xtemp[,1], y=xtemp[,2], z=xtemp[,3], type="n", xlab=xlab, ylab=ylab, zlab=zlab, ...)
-
-    for (j in 1:m)
-    {
-        for (i in 1:nhts)
-        { 
-            cti <- hts[[j]][nhts-i+1]
-            if (cti <= max(fhat$estimate[[j]]))
-                misc3d::contour3d(x=fhat$eval.points[[1]], y=fhat$eval.points[[2]], z=fhat$eval.points[[3]], f=fhat$estimate[[j]], level=cti, add=TRUE, alpha=alphavec[i], color=colors[j], ...)
+    disp1 <- match.arg(display, c("plot3D", "rgl")) 
+	if (disp1 %in% "plot3D")
+	{
+        for (j in 1:m)
+    	{
+        	for (i in 1:nhts)
+        	{ 
+            	cti <- hts[[j]][nhts-i+1]
+            	if (cti <= max(fhat$estimate[[j]]))
+                	plot3D::isosurf3D(x=fhat$eval.points[[1]], y=fhat$eval.points[[2]], z=fhat$eval.points[[3]], colvar=fhat$estimate[[j]], level=cti, add=add | (j>1 | i>1), alpha=alphavec[i], col=col[j], phi=phi, theta=theta, xlab=xlab, ylab=ylab, zlab=zlab, d=d, ticktype=ticktype, bty=bty, ...)
+        	}
         }
-        if (drawpoints)   ## plot points
+        
+        ## plot points
+        if (drawpoints)   
         {
-            if (missing(y))
-                rgl::points3d(fhat$x[[j]][,1], fhat$x[[j]][,2], fhat$x[[j]][,3], color=col.pt[j], size=size, alpha=1)
-            else
-            {
-                if (missing(y.group))
-                    rgl::points3d(y[,1], y[,2], y[,3], color=col.pt, size=size, alpha=1)
-                else
-                {
-                    y.temp <- y[y.group==levels(y.group)[j],]
-                    if (nrow(y.temp)>0)
-                        rgl::points3d(y.temp[,1], y.temp[,2], y.temp[,3], color=col.pt[j], size=size, alpha=1)
-                }
-            }
-        }
+          	for (j in 1:m)
+    		{
+            	if (missing(y))
+            	    plot3D::points3D(fhat$x[[j]][,1], fhat$x[[j]][,2], fhat$x[[j]][,3], col=col.pt[j], cex=cex, alpha=1, add=TRUE, pch=pch[j])
+            	else
+            	{
+                	if (missing(y.group))
+                    	plot3D::points3D(y[,1], y[,2], y[,3], col=col.pt, cex=cex, alpha=alpha, add=TRUE, pch=pch[j])
+                	else
+                	{
+                    	y.temp <- y[y.group==levels(y.group)[j],]
+                    	if (nrow(y.temp)>0)
+                        	plot3D::points3D(y.temp[,1], y.temp[,2], y.temp[,3], col=col.pt[j], cex=cex, alpha=alpha, add=TRUE, pch=pch[j])
+                	}
+            	}
+        	}
+    	}
+	
     }
+    else if (disp1 %in% "rgl")
+    {
+    	## suggestions from Viktor Petukhov 08/03/2018
+    	if (!requireNamespace("rgl", quietly=TRUE)) stop("Install the rgl package as it is required.", call.=FALSE)
+    	if (!requireNamespace("misc3d", quietly=TRUE)) stop("Install the misc3d package as it is required.", call.=FALSE)
+
+    	if (missing(colors)) colors <- col
+		
+		xtemp <- numeric(); for (i in 1:length(fhat$x)) xtemp <- rbind(xtemp, fhat$x[[i]])
+    	rgl::plot3d(x=xtemp[,1], y=xtemp[,2], z=xtemp[,3], type="n", xlab=xlab, ylab=ylab, zlab=zlab, ...)
+
+    	for (j in 1:m)
+    	{
+        	for (i in 1:nhts)
+        	{ 
+            	cti <- hts[[j]][nhts-i+1]
+            	if (cti <= max(fhat$estimate[[j]]))
+                	misc3d::contour3d(x=fhat$eval.points[[1]], y=fhat$eval.points[[2]], z=fhat$eval.points[[3]], f=fhat$estimate[[j]], level=cti, add=TRUE, alpha=alphavec[i], color=colors[j], ...)
+        	}
+        	if (drawpoints)   ## plot points
+        	{
+            	if (missing(y))
+            	    rgl::points3d(fhat$x[[j]][,1], fhat$x[[j]][,2], fhat$x[[j]][,3], color=col.pt[j], size=size, alpha=alpha)
+            	else
+            	{
+                	if (missing(y.group))
+                    	rgl::points3d(y[,1], y[,2], y[,3], color=col.pt, size=size, alpha=1)
+                	else
+                	{
+                    	y.temp <- y[y.group==levels(y.group)[j],]
+                    	if (nrow(y.temp)>0)
+                        	rgl::points3d(y.temp[,1], y.temp[,2], y.temp[,3], color=col.pt[j], size=size, alpha=alpha)
+                	}
+            	}
+        	}
+    	}
+    }	
 }
 
 
