@@ -77,10 +77,10 @@ kdde <- function(x, H, h, deriv.order=0, gridsize, gridtype, xmin, xmax, supp=3.
   }
   fhat$binned <- binned
   fhat$names <- parse.name(x)
+  fhat$type <- "kdde"
   class(fhat) <- "kdde"
   return(fhat)
- }
-
+}
 
 
 ###############################################################################
@@ -252,13 +252,6 @@ kdde.binned.nd <- function(H, deriv.order, bin.par, verbose=FALSE, deriv.vec=TRU
           }
       else
       {
-          ##for (s in 1:ncol(keval))
-          ##{  
-          ##    kevals <- array(keval[,s], dim=N)
-          ##    if (r==0) sf <- rep(1,d)
-          ##     else sf <- (-1)^deriv.index[s,]
-          ##    est[[s]] <- symconv.nd(kevals, bin.par$counts, d=d)
-          ##}
           kevals <- lapply(as.data.frame(keval), function(x){array(x, dim=N)})
           est <- lapply(kevals, function(x){symconv.nd(x, bin.par$counts, d=d)})
       }
@@ -272,7 +265,7 @@ kdde.binned.nd <- function(H, deriv.order, bin.par, verbose=FALSE, deriv.vec=TRU
 
 
 #############################################################################
-#### Univariate kernel density derivative estimate on a grid
+## Univariate kernel density derivative estimate on a grid
 #############################################################################
 
 kdde.grid.1d <- function(x, h, gridsize, supp=3.7, positive=FALSE, adj.positive, xmin, xmax, gridtype, w, deriv.order=0)
@@ -307,8 +300,6 @@ kdde.grid.1d <- function(x, h, gridsize, supp=3.7, positive=FALSE, adj.positive,
   return(fhatr)
 }
 
-
-
 ##############################################################################
 ## Bivariate kernel density derivative estimate on a grid
 ## Computes all mixed partial derivatives for a given deriv.order
@@ -383,6 +374,11 @@ kdde.grid.2d <- function(x, H, gridsize, supp, gridx=NULL, grid.pts=NULL, xmin,
   return(fhatr)
 }
 
+##############################################################################
+## Trivariate kernel density derivative estimate on a grid
+## Computes all mixed partial derivatives for a given deriv.order
+##############################################################################
+
 kdde.grid.3d <- function(x, H, gridsize, supp, gridx=NULL, grid.pts=NULL, xmin, xmax, gridtype, w, deriv.order=0, deriv.vec=TRUE, verbose=FALSE)
 {
   d <- 3
@@ -430,7 +426,7 @@ kdde.grid.3d <- function(x, H, gridsize, supp, gridx=NULL, grid.pts=NULL, xmin,
               for (j in 1:length(eval.y))
                   fhat.grid[[ell]][eval.x.ind,eval.y.ind[j], eval.z.ind[k]] <- 
                       fhat.grid[[ell]][eval.x.ind, eval.y.ind[j], eval.z.ind[k]] + 
-                          fhat[((j-1) * eval.x.len + 1):(j * eval.x.len)]
+                          fhat[((j-1) * eval.x.len + 1):(j * eval.x.len), ell]
           }
       
       if (verbose) setTxtProgressBar(pb, i/n)
@@ -465,7 +461,6 @@ kdde.grid.3d <- function(x, H, gridsize, supp, gridx=NULL, grid.pts=NULL, xmin,
   return(fhatr)
 }
 
-
 #############################################################################
 ## Multivariate kernel density estimate using normal kernels,
 ## evaluated at each sample point
@@ -496,9 +491,10 @@ kdde.points <- function(x, H, eval.points, w, deriv.order=0, deriv.vec=TRUE)
 }
 
 #############################################################################
-## Plot method for KDDE 
+## S3 methods for KDDE objects
 #############################################################################
 
+## plot method
 plot.kdde <- function(x, ...)
 {
   fhat <- x
@@ -527,7 +523,6 @@ plot.kdde <- function(x, ...)
       }
       else if (d==3)
       {
-        ##stop("plot.kdde not yet implemented for 3-d data")
         plotkdde.3d(fhat, ...)
         invisible()
       }
@@ -537,56 +532,77 @@ plot.kdde <- function(x, ...)
   }
 }
 
-plotkdde.1d <- function(fhat, ylab="Density derivative function", cont=50, abs.cont, ...)
+plotkdde.1d <- function(fhat, xlab, ylab="Density derivative function", cont=50, abs.cont, ...)
 {
-  if (missing(abs.cont))
-  {
-    abs.cont <- as.matrix(contourLevels(fhat, approx=TRUE, cont=cont), ncol = length(cont))
-    abs.cont <- c(abs.cont[1, ], rev(abs.cont[2, ]))
-  }
-  class(fhat) <- "kde"
+  	if (missing(xlab)) xlab <- fhat$names[1]
+    
+  	if (missing(abs.cont))
+  	{
+    	abs.cont <- as.matrix(contourLevels(fhat, approx=TRUE, cont=cont), ncol = length(cont))
+    	abs.cont <- c(abs.cont[1, ], rev(abs.cont[2, ]))
+  	}
+  	class(fhat) <- "kde"
 
-  plot(fhat, ylab=ylab, abs.cont=abs.cont, ...)
+  	plot(fhat, xlab=xlab, ylab=ylab, abs.cont=abs.cont, ...)
 }
 
 
-plotkdde.2d <- function(fhat, which.deriv.ind=1, cont=c(25,50,75), abs.cont, display="slice", zlab="Density derivative function", col.fun=topo.colors, kdde.flag=TRUE, thin=5, transf=1/4, neg.grad=FALSE, ...)
+plotkdde.2d <- function(fhat, which.deriv.ind=1, cont=c(25,50,75), abs.cont, display="slice", xlab, ylab, zlab="Density derivative function", col, col.fun, alpha=1, kdde.flag=TRUE, thin=3, transf=1, neg.grad=FALSE, ...)
 {
-  disp1 <- match.arg(display, c("persp", "slice", "image", "filled.contour", "filled.contour2", "quiver"))
-  
-  if (disp1=="slice" | disp1=="filled.contour" | disp1=="filled.contour2")
+  disp1 <- match.arg(display, c("slice", "persp", "image", "filled.contour", "filled.contour2", "quiver"))
+  if (disp1=="filled.contour2") disp1 <- "filled.contour"
+  #disp1 <- match.arg(display, c("persp", "slice", "image", "filled.contour", "quiver"))
+  if (missing(col.fun)) col.fun <- function(n) {hcl.colors(n, palette="Blue-Red", alpha=alpha)}
+  if (missing(xlab)) xlab <- fhat$names[1]
+  if (missing(ylab)) ylab <- fhat$names[2]
+  if (disp1=="slice" | disp1=="filled.contour")
   {
     if (missing(abs.cont))
     {
-      abs.cont <- as.matrix(contourLevels(fhat, approx=TRUE, cont=cont), ncol=length(cont))
+      abs.cont <- as.matrix(contourLevels(fhat, approx=TRUE, cont=cont, which.deriv.ind=which.deriv.ind), ncol=length(cont))
       abs.cont <- c(abs.cont[1,], rev(abs.cont[2,]))
     }
+    if (missing(col))
+    {
+        if (disp1=="slice") col <- col.fun(n=length(abs.cont)) 
+        else if (disp1=="filled.contour") 
+        {
+            col <- col.fun(n=length(abs.cont)+1)
+            col[median(1:length(col))] <- "transparent"
+        }
+    }
   } 
  
   if (disp1=="quiver")
   {
       if (fhat$deriv.order==1)
-          plotquiver(fhat=fhat, thin=thin, transf=transf, neg.grad=neg.grad, ...)
+          plotquiver(fhat=fhat, thin=thin, transf=transf, neg.grad=neg.grad, col=col, xlab=xlab, ylab=ylab, alpha=alpha, ...)
       else warning("Quiver plot requires gradient estimate.")
   }
   else
   {
-      fhat.temp <- fhat 
+      fhat.temp <- fhat
       fhat.temp$deriv.ind <- fhat.temp$deriv.ind[which.deriv.ind,]
       fhat.temp$estimate <- fhat.temp$estimate[[which.deriv.ind]]
       fhat <- fhat.temp
       class(fhat) <- "kde"
       
-      plot(fhat, display=display, abs.cont=abs.cont, zlab=zlab, col.fun=col.fun, kdde.flag=kdde.flag, ...) 
+      if (disp1=="persp") plot(fhat, display=display, abs.cont=abs.cont, xlab=xlab, ylab=ylab, zlab=zlab, col.fun=col.fun, kdde.flag=kdde.flag, col=col, thin=thin, alpha=alpha, ...) 
+      else plot(fhat, display=display, abs.cont=abs.cont, xlab=xlab, ylab=ylab, zlab=zlab, col.fun=col.fun, kdde.flag=kdde.flag, col=col, alpha=alpha, ...) 
   }
 }
 
 
-plotkdde.3d <- function(fhat, which.deriv.ind=1, cont=c(25,50,75), abs.cont, colors, col.fun=cm.colors, ...)
+plotkdde.3d <- function(fhat, which.deriv.ind=1, display="plot3D", cont=c(25,50,75), abs.cont, colors, col, col.fun, xlab, ylab, zlab, ...)
 {
+  if (missing(col.fun)) col.fun <- function(n) {hcl.colors(n, palette="Blue-Red")}
+  if (missing(xlab)) xlab <- fhat$names[1]
+  if (missing(ylab)) ylab <- fhat$names[2]
+  if (missing(zlab)) ylab <- fhat$names[3]
+  
   if (missing(abs.cont))
   {
-      abs.cont <- as.matrix(contourLevels(fhat, approx=TRUE, cont=cont), ncol=length(cont))
+      abs.cont <- as.matrix(contourLevels(fhat, approx=TRUE, cont=cont, which.deriv.ind=which.deriv.ind), ncol=length(cont))
       abs.cont <- c(abs.cont[1,], rev(abs.cont[2,]))
   }
   
@@ -596,13 +612,15 @@ plotkdde.3d <- function(fhat, which.deriv.ind=1, cont=c(25,50,75), abs.cont, col
   fhat <- fhat.temp
   class(fhat) <- "kde"
 
-  if (missing(colors))
+  if (missing(col))
   {
-      colors <- rev(col.fun(length(abs.cont)))
-      nc <- length(colors)
-      colors[1: (nc/2)] <- rev(colors[1:(nc/2)])
-  }
-  plot(fhat, abs.cont=abs.cont, colors=colors, ...) 
+        col <- col.fun(n=length(abs.cont)+1)
+        nc <- length(col)
+        col <- col[-median(1:nc)] 
+   }
+
+  colors <- col
+  plot(fhat, display=display, abs.cont=abs.cont, colors=col, col=col, xlab=xlab, ylab=ylab, zlab=zlab, ...) 
 }
 
 
@@ -610,11 +628,12 @@ plotkdde.3d <- function(fhat, which.deriv.ind=1, cont=c(25,50,75), abs.cont, col
 ## Quiver plot
 ######################################################################
 
-plotquiver <- function(fhat, thin=5, transf=1/4, neg.grad=FALSE, xlab, ylab, ...)
+plotquiver <- function(fhat, thin=5, transf=1, neg.grad=FALSE, xlab, ylab, col, add=FALSE, scale, length=0.1, alpha=1, ...)
 {
-    ## suggestions from Viktor Petukhov 08/03/2018
-    if (!requireNamespace("OceanView", quietly=TRUE)) stop("Install the OceanView package as it is required.", call.=FALSE)
-
+    if (!requireNamespace("pracma", quietly=TRUE)) stop("Install the pracma package as it is required.", call.=FALSE)
+    if (missing(col)) col <- 1
+    col <- transparency.col(col, alpha=alpha)
+    
     ev <- fhat$eval.points
     est <- fhat$estimate
 
@@ -625,20 +644,43 @@ plotquiver <- function(fhat, thin=5, transf=1/4, neg.grad=FALSE, xlab, ylab, ...
     
     thin1.ind <- seq(1, length(ev[[1]]), by=thin)
     thin2.ind <- seq(1, length(ev[[2]]), by=thin)
-
+    evx <- ev[[1]][thin1.ind]
+    evy <- ev[[2]][thin2.ind]
     fx <- est[[1]][thin1.ind, thin2.ind]
     fy <- est[[2]][thin1.ind, thin2.ind]
     if (neg.grad) { fx <- -fx; fy <- -fy }
     if (missing(xlab)) xlab <- fhat$names[1]
     if (missing(ylab)) ylab <- fhat$names[2]
+   
+    if (!add) plot(fhat, abs.cont=c(0,0), col="transparent", xlab=xlab, ylab=ylab)
+    grid.xy <- pracma::meshgrid(evy, evx)
+    x0 <- grid.xy$Y
+    y0 <- grid.xy$X
     
-    OceanView::quiver2D(x=ev[[1]][thin1.ind], y=ev[[2]][thin2.ind], u=fx, v=fy, xlab=xlab, ylab=ylab, ...)
+    ## default scale factor - arrows should exceed a bin
+	arrow.len <- sqrt(fx^2 + fy^2) 
+	bin.diag <- min(diff(evx), diff(evy)) 
+	if (missing(scale)) scale <- bin.diag/max(arrow.len)
+	
+	## remove `zero-length' arrows i.e. length < 1e-3 inches
+	## adapted from https://stackoverflow.com/questions/52689959/how-small-is-a-zero-length-arrow/52690054
+	units <- par(c('usr', 'pin'))
+	x2in <- with(units, pin[1L]/diff(usr[1:2]))
+	y2in <- with(units, pin[2L]/diff(usr[3:4]))
+	arrow2in <- sqrt((x2in*scale*fx)^2 + (y2in*scale*fy)^2)
+	
+	arrowint <- sort(c(min(arrow2in)-0.1*abs(max(arrow2in)), seq(1e-3, length, length=5), max(arrow2in)+0.1*abs(max(arrow2in))))
+	nsmall.f <- cut(arrow2in, arrowint, labels=FALSE)
+	nsf.lab <- unique(nsmall.f[nsmall.f>1])
+	
+	for (i in nsf.lab)
+	{
+        nsf <- which(nsmall.f==i)
+        if (length(nsf)>0) pracma::quiver(x=x0[nsf], y=y0[nsf], u=fx[nsf], v=fy[nsf], col=col, scale=scale, length=arrowint[i], ...)
+	}
 }
 
-  
-#############################################################################
-## ContourLevels method for KDDE 
-#############################################################################
+## contourLevels method 
 
 contourLevels.kdde <- function(x, prob, cont, nlevels=5, approx=TRUE, which.deriv.ind=1, ...)
 { 
@@ -700,9 +742,7 @@ contourLevels.kdde <- function(x, prob, cont, nlevels=5, approx=TRUE, which.deri
   return(hts)
 }
 
-#############################################################################
 ## predict method for KDDE 
-#############################################################################
 
 predict.kdde <- function(object, ..., x)
 {
@@ -721,7 +761,6 @@ predict.kdde <- function(object, ..., x)
     if (is.vector(fhat$deriv.ind)) pk.mat <- predict.kde(fhat, x=x, ...)
     else
     {
-     
       nd <- nrow(fhat$deriv.ind)
       pk.mat <- matrix(0, ncol=nd, nrow=n)
       for (i in 1:nd)
@@ -739,8 +778,6 @@ predict.kdde <- function(object, ..., x)
   return(drop(pk.mat))
 }
 
-
-
 ######################################################################
 ## Summary kernel curvature 
 ######################################################################
@@ -777,6 +814,7 @@ kcurv <- function(fhat, compute.cont=TRUE)
         fhat.curv$cont <- contourLevels(fhat.temp, cont=1:99)
     }
     class(fhat.curv) <- "kde"
+    fhat.curv$type <- "kcurv"
     return(fhat.curv)
 }
 
