@@ -263,17 +263,14 @@ grid.interp.3d <- function(x, gridx, f)
    return(out$est)
 }
 
-###############################################################################
 ## Linear intepolation based on kernel estimation grid
-###############################################################################
+## alias for predict.kde
 
 kde.approx <- function(fhat, x)
 {
   return(grid.interp(x=x, gridx=fhat$eval.points, f=fhat$estimate))
 }
 
-
-
 ##############################################################################
 ## Find the nearest grid points surrounding point x for non-uniform grids 
 ##############################################################################
@@ -325,7 +322,6 @@ varying.grid.interp <- function(x, gridx, f)
   return(fx)
 }
 
-
 varying.grid.interp.1d <- function(x, gridx, f)
 {
   n <- length(x)
@@ -360,36 +356,6 @@ varying.grid.interp.1d <- function(x, gridx, f)
 }
 
 
-
-predict.kde <- function(object, ..., x, zero.flag=TRUE)
-{
-  fhat <- grid.interp(x=x, gridx=object$eval.points, f=object$estimate)
-  if (!zero.flag)
-  {    
-      if (!is.list(object$eval.points)) d <- 1 else d <- length(object$eval.points)
-      if (d==1)
-      {
-          gs <- length(object$eval.points)
-          x.ind <- findInterval(x, object$eval.points, all.inside=FALSE)
-          fhat[x.ind==0] <- object$estimate[1] 
-          fhat[x.ind==gs] <- object$estimate[gs]
-      }
-      else
-      {
-          x <- matrix(x, ncol=d)
-          gs <- sapply(object$eval.points, length)
-          x.ind <- matrix(0, nrow=nrow(x), ncol=d)
-          for (i in 1:d) x.ind[,i] <- findInterval(x[,i], object$eval.points[[i]], all.inside=FALSE)
-          x.ind[x.ind==0] <- 1
-          x.ind.flag <- x.ind==1
-          for (i in 1:d) x.ind.flag[,i] <- x.ind.flag[,i] | x.ind[,i]==gs[i] 
-          fhat[apply(x.ind.flag, 1, any)] <- object$estimate[x.ind[apply(x.ind.flag, 1, any),]]
-      }
-  }
-  return(fhat)
-}
-
-
 ###############################################################################
 ## Multivariate kernel density estimate using normal kernels
 ##
@@ -497,6 +463,7 @@ kde <- function(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points,
     fhat$binned <- binned
     fhat$names <- parse.name(x)  ## add variable names
     fhat$w <- w
+    fhat$type <- "kde" 
     class(fhat) <- "kde"
   
     ## compute prob contour levels
@@ -877,13 +844,41 @@ kde.points.1d <- function(x, h, eval.points, positive=FALSE, adj.positive, w)
   return(fhat)
 }
 
+#############################################################################
+## S3 methods for KDE objects
+#############################################################################
 
-###############################################################################
-## Display kernel density estimate
-##
-## Parameters
-## fhat - output from call to `kde'
-###############################################################################
+## predict method for KDE objects
+
+predict.kde <- function(object, ..., x, zero.flag=TRUE)
+{
+  fhat <- grid.interp(x=x, gridx=object$eval.points, f=object$estimate)
+  if (!zero.flag) warning("zero.flag=FALSE has been deprecated and no longer has any effect")
+  #{    
+  #    if (!is.list(object$eval.points)) d <- 1 else d <- length(object$eval.points)
+  #    if (d==1)
+  #    {
+  #        gs <- length(object$eval.points)
+  #        x.ind <- findInterval(x, object$eval.points, all.inside=FALSE)
+  #        fhat[x.ind==0] <- object$estimate[1] 
+  #        fhat[x.ind==gs] <- object$estimate[gs]
+  #    }
+  #    else
+  #    {
+  #        x <- matrix(x, ncol=d)
+  #        gs <- sapply(object$eval.points, length)
+  #        x.ind <- matrix(0, nrow=nrow(x), ncol=d)
+  #        for (i in 1:d) x.ind[,i] <- findInterval(x[,i], object$eval.points[[i]], all.inside=FALSE)
+  #        x.ind[x.ind==0] <- 1
+  #        x.ind.flag <- x.ind==1
+  #        for (i in 1:d) x.ind.flag[,i] <- x.ind.flag[,i] | x.ind[,i]==gs[i] 
+  #        fhat[apply(x.ind.flag, 1, any)] <- object$estimate[x.ind[apply(x.ind.flag, 1, any),]]
+  #    }
+  #}
+  return(fhat)
+}
+
+## plot method
 
 plot.kde <- function(x, ...)
 { 
@@ -913,9 +908,10 @@ plot.kde <- function(x, ...)
 }
 
 plotkde.1d <- function(fhat, xlab, ylab="Density function", add=FALSE,
-  drawpoints=FALSE, col=1, col.pt="blue", col.cont=1, cont.lwd=1, jitter=FALSE, cont, abs.cont, approx.cont=TRUE, ...) 
+  drawpoints=FALSE, col=1, col.pt=4, col.cont=1, cont.lwd=1, jitter=FALSE, cont, abs.cont, approx.cont=TRUE, alpha=1, ...) 
 {
   if (missing(xlab)) xlab <- fhat$names
+  col <- transparency.col(col, alpha=alpha)
   
   if (add) lines(fhat$eval.points, fhat$estimate, xlab=xlab, ylab=ylab, col=col, ...)
   else plot(fhat$eval.points, fhat$estimate, type="l", xlab=xlab, ylab=ylab, col=col, ...) 
@@ -927,11 +923,14 @@ plotkde.1d <- function(fhat, xlab, ylab="Density function", add=FALSE,
     {
       if (!is.null(fhat$cont))
       {
-        cont.ind <- rep(FALSE, length(fhat$cont))
-        for (j in 1:length(cont))
-          cont.ind[which(cont[j] == 100-as.numeric(unlist(strsplit(names(fhat$cont),"%"))))] <- TRUE
-
-        if (all(!cont.ind))
+          cont.ind <- rep(FALSE, length(cont))
+          for (j in 1:length(cont))
+          {
+              ci <- which(cont[j]==(100-as.numeric(unlist(strsplit(names(fhat$cont),"%")))))
+              if (length(ci)==0) cont.ind[j] <- NA else cont.ind[j] = ci
+          }
+          
+        if (any(is.na(cont.ind)))
           hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
         else
           hts <- fhat$cont[cont.ind]
@@ -941,7 +940,6 @@ plotkde.1d <- function(fhat, xlab, ylab="Density function", add=FALSE,
     }
     else
       hts <- abs.cont 
-
     ##
     
     if (is.null(fhat$deriv.order))
@@ -963,39 +961,32 @@ plotkde.1d <- function(fhat, xlab, ylab="Density function", add=FALSE,
   }
   
   if (drawpoints)
-    if (jitter)
-      rug(jitter(fhat$x), col=col.pt)
-    else
-      rug(fhat$x, col=col.pt)
+    if (jitter) rug(jitter(fhat$x), col=col.pt)
+    else rug(fhat$x, col=col.pt)
 }
 
-
-###############################################################################
-## Display bivariate kernel density estimate
-##
-## Parameters 
-## fhat - output from 'kde.grid'
-## display - "persp" - perspective plot
-##         - "slice" - contour plot
-##         - "image" image plot
-## cont - vector of contours to be plotted
-###############################################################################
-
-plotkde.2d <- function(fhat, display="slice", cont=c(25,50,75), abs.cont, approx.cont=TRUE, xlab, ylab, zlab="Density function", cex=1, pch=1, add=FALSE, drawpoints=FALSE, drawlabels=TRUE, theta=-30, phi=40, d=4, col.pt="blue", col, col.fun, lwd=1, border=1, thin=3, lwd.fc=5, kdde.flag=FALSE, ...) 
+plotkde.2d <- function(fhat, display="slice", cont=c(25,50,75), abs.cont, approx.cont=TRUE, xlab, ylab, zlab="Density function", cex=1, pch=1, labcex=1, add=FALSE, drawpoints=FALSE, drawlabels=TRUE, theta=-30, phi=40, d=4, col.pt=4, col, col.fun, alpha=1, lwd=1, border=1, thin=3, kdde.flag=FALSE, ticktype="detailed", ...) 
 {
   disp1 <- match.arg(display, c("slice", "persp", "image", "filled.contour", "filled.contour2"))
+  if (disp1=="filled.contour2") disp1 <- "filled.contour"
+  ## disp=="filled.contour2" is deprecated but is still used in simukde package
   if (!is.list(fhat$eval.points)) stop("Need a grid of density estimates")
 
   if (missing(xlab)) xlab <- fhat$names[1]
   if (missing(ylab)) ylab <- fhat$names[2]
-
+  if (missing(col.fun)) 
+  {
+    if (any(fhat$type=="kcurv")) col.fun <- function(n) {hcl.colors(n, palette="Oranges",rev=TRUE, alpha=alpha) }
+    else col.fun <- function(n) {hcl.colors(n, palette="heat",rev=TRUE, alpha=alpha)}
+  }
+  
   ## perspective/wireframe plot
   if (disp1=="persp")
   {
-    hts <- seq(0, 1.1*max(fhat$estimate,na.rm=TRUE), length=500)
-    if (missing(col)) col <- topo.colors(length(hts)+1, alpha=0.5)
-    if (!missing(col.fun)) col <- col.fun(length(hts)+1)
+    hts <- seq(0, 1.1*max(fhat$estimate,na.rm=TRUE), length=100)
+    if (missing(col)) col <- col.fun(length(hts)+1)
     if (length(col)<length(hts)) col <- rep(col, length=length(hts))
+    col <- transparency.col(col, alpha=alpha)
     
     ## thinning indices
     plot.ind <- list(seq(1, length(fhat$eval.points[[1]]), by=thin), seq(1, length(fhat$eval.points[[2]]), by=thin))
@@ -1006,7 +997,7 @@ plotkde.2d <- function(fhat, display="slice", cont=c(25,50,75), abs.cont, approx
     zfacet <- z[-1, -1] + z[-1, -ncz] + z[-nrz, -1] + z[-nrz, -ncz]
     facetcol <- cut(zfacet, length(col)+1)
     
-    plotret <- persp(fhat$eval.points[[1]][plot.ind[[1]]], fhat$eval.points[[2]][plot.ind[[2]]], z, theta=theta, phi=phi, d=d, xlab=xlab, ylab=ylab, zlab=zlab, col=col[facetcol], border=border, ...)
+    plotret <- persp(fhat$eval.points[[1]][plot.ind[[1]]], fhat$eval.points[[2]][plot.ind[[2]]], z, theta=theta, phi=phi, d=d, xlab=xlab, ylab=ylab, zlab=zlab, col=col[facetcol], border=border, ticktype=ticktype, ...)
   }
   else if (disp1=="slice") 
   {
@@ -1015,11 +1006,14 @@ plotkde.2d <- function(fhat, display="slice", cont=c(25,50,75), abs.cont, approx
     {
       if (!is.null(fhat$cont))
       {
-        cont.ind <- rep(FALSE, length(fhat$cont))
-        for (j in 1:length(cont))
-          cont.ind[which(cont[j] == 100-as.numeric(unlist(strsplit(names(fhat$cont),"%"))))] <- TRUE
-        
-        if (all(!cont.ind))
+          cont.ind <- rep(FALSE, length(cont))
+          for (j in 1:length(cont))
+          {
+              ci <- which(cont[j]==(100-as.numeric(unlist(strsplit(names(fhat$cont),"%")))))
+              if (length(ci)==0) cont.ind[j] <- NA else cont.ind[j] = ci
+          }
+          
+        if (any(is.na(cont.ind)))
           hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
         else
           hts <- fhat$cont[cont.ind]
@@ -1030,20 +1024,26 @@ plotkde.2d <- function(fhat, display="slice", cont=c(25,50,75), abs.cont, approx
     else
       hts <- abs.cont 
     
-    if (missing(col)) col <- 1 #rev(heat.colors(length(hts)))
+    hts <- sort(hts)
+    if (missing(col)) col <- col.fun(length(hts))
     if (length(col)<length(hts)) col <- rep(col, times=length(hts))
-
+    col <- transparency.col(col, alpha=alpha)
+    
     ## draw contours
     j <- 0
     for (i in 1:length(hts)) 
     {
-      if (missing(abs.cont)) scale <- cont[i]/hts[i]
-      else scale <- 1
-
+      if (missing(abs.cont)) { ni <- length(hts)-i+1; scale <- (100-cont[i])/hts[i]; scale2 <- cont[ni]/hts[ni] }
+      else { ni <- i; scale <- 1; scale2 <-1 }
+    
+      
       if (hts[i]>0 | !is.null(fhat$deriv.order))
       {
           j<-j+1;
-          contour(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate*scale, level=hts[i]*scale, add=j>1 | add, drawlabels=drawlabels, col=col[i], lwd=lwd, xlab=xlab, ylab=ylab, ...)
+          if (j==1) 
+          	contour(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate*scale, level=hts[i]*scale, label=signif(hts[ni]*scale2), add=add, drawlabels=drawlabels, col=col[i], lwd=lwd, labcex=labcex, xlab=xlab, ylab=ylab, ...)
+          else 
+          	contour(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate*scale, level=hts[i]*scale, label=signif(hts[ni]*scale2), add=TRUE, drawlabels=drawlabels, col=col[i], lwd=lwd, labcex=labcex, ...)
       }
     }
  
@@ -1053,122 +1053,100 @@ plotkde.2d <- function(fhat, display="slice", cont=c(25,50,75), abs.cont, approx
   ## image plot
   else if (disp1=="image")
   {
-    if (missing(col)) col <- rev(heat.colors(100))  
+    if (missing(col)) col <- col.fun(100)
+    col <- transparency.col(col, alpha=alpha)
     image(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate, xlab=xlab, ylab=ylab, add=add, col=col, ...)
 
     ## add points 
     if (drawpoints) points(fhat$x[,1], fhat$x[,2], col=col.pt, cex=cex, pch=pch)
     box()
   }
-  else if (disp1=="filled.contour" | disp1=="filled.contour2")
+  else if (disp1=="filled.contour")
   {
     ## compute contours
     if (missing(abs.cont))
     {
       if (!is.null(fhat$cont))
       {
-        cont.ind <- rep(FALSE, length(fhat$cont))
-        for (j in 1:length(cont))
-          cont.ind[which(cont[j] == 100-as.numeric(unlist(strsplit(names(fhat$cont),"%"))))] <- TRUE
-        
-        if (all(!cont.ind))
+          cont.ind <- rep(FALSE, length(cont))
+          for (j in 1:length(cont))
+          {
+              ci <- which(cont[j]==(100-as.numeric(unlist(strsplit(names(fhat$cont),"%")))))
+              if (length(ci)==0) cont.ind[j] <- NA else cont.ind[j] = ci
+          }
+          
+        if (any(is.na(cont.ind)))
           hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
         else
           hts <- fhat$cont[cont.ind]
       }
       else
         hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
-    }  
+    }
     else
       hts <- abs.cont 
     hts <- sort(hts)
-
-    if (missing(col)) col <- c("transparent", rev(heat.colors(length(hts))))
-    if (!missing(col.fun)) col <- c(col.fun(length(hts)+1))
+    
+    if (missing(col)) col <- c("transparent", col.fun(length(hts)))
+    col <- transparency.col(col, alpha=alpha)
     clev <- c(min(c(fhat$estimate, hts)-0.01*max(abs(fhat$estimate))), hts, max(c(fhat$estimate, hts)) + 0.01*max(abs(fhat$estimate)))
+
+    if (!add) plot(fhat$eval.points[[1]], fhat$eval.points[[2]], type="n", xlab=xlab, ylab=ylab, ...)
+    .filled.contour(fhat$eval.points[[1]], fhat$eval.points[[2]], z=fhat$estimate, levels=clev, col=col)
       
-    if (disp1=="filled.contour2")
+    if (!missing(lwd))
     {
-        image(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate, xlab=xlab, ylab=ylab, add=add, col=col[1:(length(hts)+1)], breaks=clev, ...)
-
-      ## draw contours         
-     
-      for (i in 1:length(hts))
-      {      
-          if (!kdde.flag) contour(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate, level=hts[i], add=TRUE, drawlabels=FALSE, col=col[i+1], lwd=lwd.fc)
-          else
-          {
-              if (i <= length(hts)%/%2)
-                  contour(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate, level=hts[i], add=TRUE, drawlabels=FALSE, col=col[i], lwd=lwd.fc)
-              else
-                  contour(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate, level=hts[i], add=TRUE, drawlabels=FALSE, col=col[i+1], lwd=lwd.fc)
-          }
-      }
+        for (i in 1:length(hts)) 
+        {
+            if (missing(abs.cont)) { ni <- length(hts)-i+1; scale <- (100-cont[i])/hts[i]; scale2 <- cont[ni]/hts[ni] }
+            else { ni <- i; scale <- 1; scale2 <-1 }
+    
+            if (lwd >=1) contour(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate*scale, level=hts[i]*scale, label=signif(hts[ni]*scale2,3), add=TRUE, drawlabels=drawlabels, col=1, lwd=lwd, labcex=labcex, ...)
+        }
     }
-    else
-    {
-        if (!add) plot(fhat$eval.points[[1]], fhat$eval.points[[2]], type="n", xlab=xlab, ylab=ylab, ...)
-        ##if (col[1]!="transparent") 
-        .filled.contour(fhat$eval.points[[1]], fhat$eval.points[[2]], z=fhat$estimate, levels=clev, col=col)
-    }
-      
-      if (!missing(lwd))
-      {
-          for (i in 1:length(hts)) 
-          {
-              if (missing(abs.cont)) scale <- cont[i]/hts[i]
-              else scale <- 1
 
-              if (lwd >=1) contour(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate*scale, level=hts[i]*scale, add=TRUE, drawlabels=drawlabels, col=1, lwd=lwd, ...)
-          }
-      }
-
-      ## add points 
-      if (drawpoints) points(fhat$x[,1], fhat$x[,2], col=col.pt, cex=cex, pch=pch)
-      box()
+    ## add points 
+    if (drawpoints) points(fhat$x[,1], fhat$x[,2], col=col.pt, cex=cex, pch=pch)
+    box()
   }
   
   if (disp1=="persp")  invisible(plotret)
   else invisible()
 }
   
-
-
-###############################################################################
-## Display trivariate kernel density estimate
-###############################################################################
-
-
-plotkde.3d <- function(fhat, cont=c(25,50,75), abs.cont, approx.cont=TRUE, colors, col.fun, alphavec, size=3, col.pt="blue", add=FALSE, xlab, ylab, zlab, drawpoints=FALSE, alpha=1, box=TRUE, axes=TRUE, ...)
+plotkde.3d <- function(fhat, display="plot3D", cont=c(25,50,75), abs.cont, approx.cont=TRUE, colors, col, col.fun, alphavec, size=3, cex=1, pch=1, theta=-30, phi=40, d=4, ticktype="detailed", bty="f", col.pt=4, add=FALSE, xlab, ylab, zlab, drawpoints=FALSE, alpha, box=TRUE, axes=TRUE, ...)
 {
-    ## suggestions from Viktor Petukhov 08/03/2018
-    if (!requireNamespace("rgl", quietly=TRUE)) stop("Install the rgl package as it is required.", call.=FALSE)
-    if (!requireNamespace("misc3d", quietly=TRUE)) stop("Install the misc3d package as it is required.", call.=FALSE)
-    
     ## compute contours
     if (missing(abs.cont))
     {
-        if (!is.null(fhat$cont))
-        {
-            cont.ind <- rep(FALSE, length(fhat$cont))
-            for (j in 1:length(cont))
-                cont.ind[which(cont[j] == 100-as.numeric(unlist(strsplit(names(fhat$cont),"%"))))] <- TRUE
-            
-            if (all(!cont.ind))
-                hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
-            else
-                hts <- fhat$cont[cont.ind]
-        }
-        else
-            hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
+       	if (!is.null(fhat$cont))
+       	{
+           	cont.ind <- rep(FALSE, length(fhat$cont))
+           	for (j in 1:length(cont))
+           	    cont.ind[which(cont[j] == 100-as.numeric(unlist(strsplit(names(fhat$cont),"%"))))] <- TRUE
+           
+	        if (all(!cont.ind))
+	            hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
+	        else
+	            hts <- fhat$cont[cont.ind]
+	     }
+	     else
+ 	        hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
     }  
     else
         hts <- abs.cont
-    
+	    
     nc <- length(hts)
-    
-    if (missing(colors)) colors <- rev(heat.colors(nc))
-    if (!missing(col.fun)) colors <- col.fun(nc)
+	if (missing(col)) 
+    { 
+        if (missing(col.fun)) 
+        {
+            if (any(fhat$type=="kcurv")) col.fun <- function(n) {hcl.colors(n, palette="Oranges",rev=TRUE) }
+            else col.fun <- function(n) {hcl.colors(n, palette="heat",rev=TRUE)}
+        }
+        col <- col.fun(n=length(hts))
+    }    
+    colors <- col
     if (missing(xlab)) xlab <- fhat$names[1]
     if (missing(ylab)) ylab <- fhat$names[2]
     if (missing(zlab)) zlab <- fhat$names[3]
@@ -1177,27 +1155,42 @@ plotkde.3d <- function(fhat, cont=c(25,50,75), abs.cont, approx.cont=TRUE, color
         if (is.null(fhat$deriv.order)) alphavec <- seq(0.1,0.5,length=nc)
         else alphavec <- c(rev(seq(0.1,0.4,length=round(nc/2))), seq(0.1,0.4,length=round(nc/2)))
     }
+    if (missing(alpha)) alpha <- 0.1 
+    else if (!missing(alpha)) {alphavec <- rep(alpha,nc)}
+    	    
+ 	disp1 <- match.arg(display, c("plot3D", "rgl")) 
+	if (disp1 %in% "plot3D")
+	{
+		 
+		for (i in 1:nc)
+    	    if (hts[nc-i+1] < max(fhat$estimate))
+    	        plot3D::isosurf3D(x=fhat$eval.points[[1]], y=fhat$eval.points[[2]], z=fhat$eval.points[[3]], colvar=fhat$estimate, level=hts[nc-i+1], add=add | (i>1), col=colors[i], alpha=alphavec[i], phi=phi, theta=theta, xlab=xlab, ylab=ylab, zlab=zlab, d=d, ticktype=ticktype, bty=bty, ...)
+    	        
+    	if (drawpoints) plot3D::points3D(x=fhat$x[,1], y=fhat$x[,2], z=fhat$x[,3], cex=cex, col=col.pt, add=TRUE, pch=pch, d=d)    
+	}
+	else if (disp1 %in% "rgl")
+	{	
+    	## suggestions from Viktor Petukhov 08/03/2018
+    	if (!requireNamespace("rgl", quietly=TRUE)) stop("Install the rgl package as it is required.", call.=FALSE)
+    	if (!requireNamespace("misc3d", quietly=TRUE)) stop("Install the misc3d package as it is required.", call.=FALSE)
     
-    fhat.eval.mean <- sapply(fhat$eval.points, mean)
-    if (drawpoints)
-        rgl::plot3d(fhat$x[,1],fhat$x[,2],fhat$x[,3], size=size, col=col.pt, alpha=alpha, xlab=xlab, ylab=ylab, zlab=zlab, add=add, box=FALSE, axes=FALSE, ...)
-    else
-        rgl::plot3d(fhat$x[,1],fhat$x[,2],fhat$x[,3], size=0, col="transparent", alpha=0, xlab=xlab, ylab=ylab, zlab=zlab, add=add, box=FALSE, axes=FALSE, ...)
+    	fhat.eval.mean <- sapply(fhat$eval.points, mean)
+	    if (drawpoints)
+	        rgl::plot3d(fhat$x[,1],fhat$x[,2],fhat$x[,3], size=size, col=col.pt, alpha=alpha, xlab=xlab, ylab=ylab, zlab=zlab, add=add, box=FALSE, axes=FALSE, ...)
+    	else
+    	    rgl::plot3d(fhat$x[,1],fhat$x[,2],fhat$x[,3], size=0, col="transparent", alpha=0, xlab=xlab, ylab=ylab, zlab=zlab, add=add, box=FALSE, axes=FALSE, ...)
 
-    for (i in 1:nc)
-        if (hts[nc-i+1] < max(fhat$estimate))
-            misc3d::contour3d(fhat$estimate, level=hts[nc-i+1], x=fhat$eval.points[[1]], y=fhat$eval.points[[2]], z=fhat$eval.points[[3]], add=TRUE, color=colors[i], alpha=alphavec[i], box=FALSE, axes=FALSE, ...)
+    	for (i in 1:nc)
+    	    if (hts[nc-i+1] < max(fhat$estimate))
+    	        misc3d::contour3d(fhat$estimate, level=hts[nc-i+1], x=fhat$eval.points[[1]], y=fhat$eval.points[[2]], z=fhat$eval.points[[3]], add=TRUE, color=colors[i], alpha=alphavec[i], box=FALSE, axes=FALSE, ...)
     
-    if (axes) rgl::axes3d()
-    if (box) rgl::box3d()
+    	if (axes) rgl::axes3d(c("x","y","z"))
+    	if (box) rgl::box3d()
+    }
 }
 
 
-
-
-###############################################################################
-## Contour levels S3 method
-###############################################################################
+## contourLevels method
 
 ## create S3 generic 
 contourLevels <- function(x, ...){  
@@ -1252,22 +1245,48 @@ contourLevels.kde <- function(x, prob, cont, nlevels=5, approx=TRUE, ...)
 
 contourSizes <- function(x, abs.cont, cont=c(25,50,75), approx=TRUE)
 {
-  num.int <- vector()
-  if (missing(abs.cont))
-    abs.cont <- contourLevels(x, cont=cont, approx=approx)
-
-  num.int <- rep(0, length(abs.cont))
-  if (!is.null(names(abs.cont))) names(num.int) <- names(abs.cont) 
-  ##abs.cont <- sort(abs.cont)
-  if (!is.list(x$eval.points))
-    delta.int <- head(diff(x$eval.points), n=1)
-  else
-    delta.int <- prod(sapply(x$eval.points, diff)[1,]) 
-  
-  for (j in 1:length(abs.cont)) 
-    num.int[j] <- sum(x$estimate>abs.cont[j])
-  
-  return(num.int*delta.int)
+    if (missing(abs.cont))
+        abs.cont <- contourLevels(x, cont=cont, approx=approx)
+    num.int <- rep(0, length(abs.cont))
+    if (!is.null(names(abs.cont))) names(num.int) <- names(abs.cont) 
+    
+    if (!is.list(x$eval.points))
+        delta.int <- head(diff(x$eval.points), n=1)
+    else
+        delta.int <- prod(sapply(lapply(x$eval.points, diff), head, n=1)) 
+    
+    for (j in 1:length(abs.cont)) 
+        num.int[j] <- sum(x$estimate>abs.cont[j])
+    
+    return(num.int*delta.int)
 }
 
 
+###############################################################################
+## Riemann sums to compute approximate probability of contour set
+###############################################################################
+
+contourProbs <- function(x, abs.cont, cont=c(25,50,75), approx=TRUE)
+{
+    if (missing(abs.cont)) 
+        abs.cont <- contourLevels(x, cont=cont, approx=approx)
+    num.int <- rep(0, length(abs.cont))
+    if (!is.null(names(abs.cont))) 
+        names(num.int) <- names(abs.cont)
+    if (!is.list(x$eval.points)) 
+    {
+        delta.int <- head(diff(x$eval.points), n=1)
+        eval.points.midpoint <- (head(x$eval.points,n=-1)+tail(x$eval.points,n=-1))/2
+    }
+    else
+    {
+        delta.int <- prod(sapply(lapply(x$eval.points, diff), head, n=1)) 
+        eval.points.midpoint <- expand.grid(lapply(x$eval.points, function(y){(head(y,n=-1)+tail(y,n=-1))/2}))
+    }
+    
+    x.evmp <- predict(x, x=eval.points.midpoint)
+    for (i in 1:length(num.int)) num.int[i] <- sum(x.evmp*(x.evmp>=abs.cont[i])*delta.int)
+    
+    return(num.int)
+}
+ 
